/**************************************************************************
   Author:         Andrea Arbore
   First release:  1.0 - 1 August 2004
   Last release:   2.0 - 29 April 2005
   Purpose:        Graph Coloring
   web site:       http://studenti.ing.uniroma1.it/~arborean  
   email:          arborean@hotmail.com

   Copyright (C) 2004-2005 by Andrea Arbore. All rights reserved.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   Permission to use, copy, modify, and distribute this software for any
   purpose without fee is hereby granted, provided that this entire notice
   is included in all copies of any software which is or includes a copy
   or modification of this software and in all copies of the supporting
   documentation for such software.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <error.h>
#include <limits.h>
#include <math.h>
#include <malloc.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/resource.h>
#define N 10000  //maximum number of vertices handles
#define LEN 25  //file name length: 20+1+3+1

volatile sig_atomic_t run=1;//global atomic variable for the signal handler

void handler_sigint(int sig)
{run=0;
 puts("\n\n***************************************************************");
     puts("*   Please, wait a moment to complete the calculations and    *");
     puts("*   to save the better partial solution in the .slt file.     *");
     puts("***************************************************************\n");
} 

typedef struct NODO //record vettore dei nodi
 { unsigned int nodo;//0..n-1
   int deg;
   int degrel; //grado relativo rispetto alla classe attuale
   int degsup; //grado relativo rispetto alle classi superiori
   int colore; //colore del nodo
   int* succ;  //puntatore al vettore dei successori
 } NODO;
 
void ordina(int **V,int d,int ind)
{int *swp,inv,i;
 do{inv=0;
    for(i=0;i<d-1;i++)
     if(V[i][ind]<V[i+1][ind])
      {swp=V[i];
       V[i]=V[i+1];
       V[i+1]=swp;
       inv++;
      }
    }
 while(inv!=0);
 return;
}

int colora(NODO *T,int nodi,int h)
 {int i,j,col=1;
  int ind,acp,deg;
  ind=h;
  do
  {T[ind].colore=col=-T[ind].colore;//inverti il segno della colorazione del nodo ind
   for(j=0;j<T[ind].deg;j++) //per tutti i successori di ind
    if(T[T[ind].succ[j]].colore==-col)//se il successore e' dello stesso livello di colore di ind
     T[T[ind].succ[j]].colore--;//coloralo negativamente rinviandolo al ciclo successivo: -col -> -(col+1)
   //Una volta colorati e rinviati tutti i successori del nodo ind
   //individua il prossimo nodo da colorare nella classe attuale -col
   ind=nodi;deg=nodi;acp=-nodi;
   for(j=0;j<nodi;j++)
    if(T[j].colore==-col)
     {T[j].degrel=0;T[j].degsup=0;
      for(i=0;i<T[j].deg;i++) //per tutti i successori di i
       {if(T[T[j].succ[i]].colore==-(col+1)) T[j].degsup++;//se il successore e' colorato per la classe successiva
        else if(T[T[j].succ[i]].colore==-col) T[j].degrel++;//se il successore e' colorato per la classe attuale
       }
      if((T[j].degrel-T[j].degsup)<deg)
       {ind=j;deg=T[j].degrel-T[j].degsup;acp=T[j].deg-T[j].degrel-T[j].degsup;}
       else
        if(((T[j].degrel-T[j].degsup)==deg)&&(acp<(T[j].deg-T[j].degrel-T[j].degsup)))
         {ind=j;acp=T[j].deg-T[j].degrel-T[j].degsup;}
     }
   if(ind==nodi)//se non vi sono nodi non colorati nella classe -col passa a quella successiva -(col+1)
    {//le inizializzazioni dei parametri di riferimento non sono state alterate
     for(j=0;j<nodi;j++)
     {if(T[j].colore==-(col+1))
      {T[j].degrel=0;
       for(i=0;i<T[j].deg;i++) //per tutti i successori di i
         if(T[T[j].succ[i]].colore==-(col+1)) T[j].degrel++;
       if(acp<T[j].degrel)
        {ind=j;acp=T[j].degrel;}         
      }
     }//fine ciclo for di scansione lista dei nodi
    }
  }while(ind<nodi); 
 return col;
 }

//\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\
/*******************************************************************/
/////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{ int i,j,k,col,colore,fine,err=0,maxg,ming,ncl;
  int a,b,maxcol;
  unsigned int n,m,s;
  sigset_t block_signal;
  float medg;
  double medcol,varcol;
  char ins,ch,form[5];
  char infile[LEN],outfile[LEN];
  FILE *file_in,*file_out;
  struct rusage tempo;
  struct timeval ttemp;
  long int tot_sec0,tot_microsec0,prg_sec0,prg_microsec0,sys_sec0,sys_microsec0;
  long int tot_sec,tot_microsec,prg_sec,prg_microsec,sys_sec,sys_microsec;
  short **M;
  int **L,*S; 
  NODO *T;

  //******** Inizio test sintattico dell'input **********
  if(argc<2)
    {printf("Usage: arbore <filename>\n");exit(EXIT_FAILURE);}
  for(i=0;i<LEN; i++) infile[i]=argv[1][i];
  file_in=fopen(infile, "r");
  if(file_in==NULL)
    {printf("ERROR: The specified file doesn't exist!\n");exit(EXIT_FAILURE);}
  while((fscanf(file_in, "%c", &ch)!=EOF)&&(ch!='p'))
    while(ch!='\n') fscanf(file_in, "%c", &ch);
  if(ch!='p'){printf("ERROR: The input file has not a problem line %cp edge <nodes> <edges>%c!\n",34,34);
              exit(EXIT_FAILURE);}
  fscanf(file_in,"%s%d%d",form,&n,&m);
  while(ch!='\n') fscanf(file_in, "%c", &ch);
  while((fscanf(file_in, "%c", &ch)!=EOF)&&(ch!='e'))
    while(ch!='\n') fscanf(file_in, "%c", &ch);
  if(strcmp(form,"edge"))
    {printf("ERROR: The problem line specifies a different format from the supported one!\n");
     printf("       The format is %cp %s%c rather than %cp edge%c as required by the DIMACS standard.\n",34,form,34,34,34);
     fclose(file_in);exit(EXIT_FAILURE);}
  if(n<1||n>N)
    {printf("ERROR: The number of vertices is out of the admitted range [1..%d].\n",N);
     fclose(file_in);exit(EXIT_FAILURE);}
  printf("\n******************************************************************************\n");    
    printf("* A simple variant of the RLF algorithm (nRLF v2.0) written by Andrea Arbore *\n");
    printf("* Copyright (C) 2004-2005 by Andrea Arbore. All rights reserved.             *\n");  
    printf("******************************************************************************\n");      
  printf("OK! The File %s has been open and its header is correct.\n",infile);
  //********** Fine test sintattico dell'input **************

  //********** Inizio elaborazione del file dati ***************
  printf("Number of DECLARED vertices: %d\n",n);
  printf("Number of DECLARED edges:    %d\n",m);
  M=(short**)malloc(n*sizeof(short*));
  if(M==NULL)
    {perror("ERROR: The virtual Memory is exhausted to allocate the adjacency matrix!\n");
     exit(EXIT_FAILURE);}
  L=(int**)malloc(n*sizeof(int));
  if(L==NULL)
    {perror("ERROR: The virtual Memory is exhausted to allocate the nodes-colors vector!\n");
     free(M);exit(EXIT_FAILURE);}
  for(k=0;k<n;k++)
   {M[k]=(short*)calloc(n,sizeof(short));
    L[k]=(int*)calloc(4,sizeof(int));
    L[k][0]=k;
   }
 ch='e';k=0;ncl=0;
 do
  {fscanf(file_in,"%d%d",&a,&b);k++;
   if((0<a)&&(a<=n)&&(0<b)&&(b<=n)&&(!M[a-1][b-1])) M[a-1][b-1]=M[b-1][a-1]=1;
    else
    {if((0<a)&&(a<=n)&&(0<b)&&(b<=n)) ncl++;
     else
      {printf("ERROR: Some vertex indexes are out of the limits [1..%d]. Row %d of the file.\n",n,k); 
       for(j=0;j<n;j++) {free(M[j]);free(L[j]);}
       free(M);free(L);fclose(file_in);exit(EXIT_FAILURE);
      }
    }
    while(ch!='\n') fscanf(file_in, "%c", &ch);
     if(fscanf(file_in, "%c", &ch)==EOF) ch='z';
  }while(ch!='z');
  for(k=0;k<n;k++)
   for(j=0;j<n;j++)
     L[k][1]+=M[k][j];
 fclose(file_in);
 if(ncl)
  {printf("WARNING: In the instance file there are %d edges duplicates! They will be neglected.",ncl);
   fflush(stdout);sleep(1);
   printf("\r                                                                                      ",ncl);
   fflush(stdout);sleep(1);
   printf("\rWARNING: In the instance file there are %d edges duplicates! They will be neglected.",ncl);
   fflush(stdout);sleep(1);
   printf("\r                                                                                      ");
   fflush(stdout);sleep(1);
   printf("\rWARNING: In the instance file there are %d edges duplicates! They will be neglected.",ncl);
   fflush(stdout);sleep(3);printf("\n");
  } 
 printf("OK! File %s has been read completely.\n",infile);
 printf("Number of READ vertices: %d\n",k);
 //*********************************************************************
 maxg=-1;ming=n;m=0;
 for(i=0;i<n;i++)
  {m+=L[i][1];
   if(L[i][1]>maxg) maxg=L[i][1];
   if(L[i][1]<ming) ming=L[i][1];
  }
 medg=(float)m/n;m=m/2;
 printf("Number of READ edges:    %d\n",m);
 printf("Maximum degree: %d\t Minimum degree: %d\t Average degree: %2.3f\n",maxg,ming,medg);
 printf("Graph Density: %1.3f\n",(2.0*m)/(n*(n-1)));
 printf("Running...\n");
 //************************************************************************//
 ordina(L,n,1);
 //************************************************************************//
 T=(NODO*)malloc(n*sizeof(NODO));
 for(k=0;k<n;k++)
  {T[k].nodo=L[k][0];
   T[k].deg=L[k][1];
   T[k].degrel=T[k].degsup=0;
   T[k].colore=-1;
   T[k].succ=(int*)calloc(L[k][1],sizeof(int));
   for(j=i=0;j<n;j++)
    if(M[L[k][0]][L[j][0]])
     T[k].succ[i++]=j;
  } 
 for(i=0;i<n;i++) free(M[i]);
 free(M);
 //********************************************************************************
 signal(SIGINT,handler_sigint);
 sigfillset(&block_signal);
 sigprocmask(SIG_BLOCK,&block_signal,NULL);
 gettimeofday(&ttemp,NULL);
 getrusage(RUSAGE_SELF,&tempo);
 tot_sec0=ttemp.tv_sec;           tot_microsec0=ttemp.tv_usec;
 prg_sec0=tempo.ru_utime.tv_sec;  prg_microsec0=tempo.ru_utime.tv_usec;
 sys_sec0=tempo.ru_stime.tv_sec;  sys_microsec0=tempo.ru_stime.tv_usec;
 colore=n;
 for(k=0;(k<n)&&run;k++) 
  {sigprocmask(SIG_UNBLOCK,&block_signal,NULL);
   col=colora(T,n,k);
   L[k][3]=col;
   printf("%d \r",n-k);fflush(stdout); 
   if(col<colore)
    {colore=col;
     printf("%d \tNew minimum coloring: %d \n",n-k,colore);
     gettimeofday(&ttemp,NULL); 
     getrusage(RUSAGE_SELF,&tempo);
     tot_sec=ttemp.tv_sec-tot_sec0;
     prg_sec= tempo.ru_utime.tv_sec-prg_sec0;
     sys_sec= tempo.ru_stime.tv_sec-sys_sec0;
     tot_microsec=ttemp.tv_usec-tot_microsec0;
     prg_microsec=tempo.ru_utime.tv_usec-prg_microsec0;
     sys_microsec=tempo.ru_stime.tv_usec-sys_microsec0;
     printf("   \tTot: %5.3f sec   CPU: %5.3f sec   Sys: %5.3f sec\n",\
     tot_sec+(tot_microsec/1E6),prg_sec+(prg_microsec/1E6),sys_sec+(sys_microsec/1E6));
     for(j=0;j<n;j++) L[j][2]=T[j].colore;
    }
   for(j=0;j<n;j++) T[j].colore=-1;
   sigprocmask(SIG_BLOCK,&block_signal,NULL);
  } 
 printf("*******************************************************************************\n");  
 printf("The best coloring uses %d colors. \n",colore);
 printf("Verification of the %d-coloring.... ",colore);
 fine=1;
 for(j=0;j<n;j++)
  for(i=0;i<T[j].deg;i++)
   if(L[j][2]==L[T[j].succ[i]][2])
    {printf("\nThe %d node is adjacent to the %d node and both they have the same %d color.",\
               L[j][1]+1,L[T[j].succ[i]][1]+1,L[j][2]);
     fine=0;
    }
 if(fine) printf("Ok! The %d-coloring is correct!\n",colore);
  else printf("\nKo! The %d-coloring is NOT correct!\n",colore);
 medcol=varcol=0.0;
 maxcol=colore;
 S=(int*)calloc(colore,sizeof(int));
 for(j=0;j<n;j++) {medcol+=L[j][3];S[L[j][2]-1]++;if(L[j][3]>maxcol) maxcol=L[j][3];}
 medcol=medcol/n; 
 for(j=0;j<n;j++) varcol+=pow(medcol-L[j][3],2);
 if(n-1) varcol=sqrt(varcol/(n-1));
 if(run) printf("Average coloring: %2.3f  Sample standard deviation: %2.3f  ",medcol,varcol);
 printf("Max coloring: %d\n",maxcol);
 printf("Classes: ");
 i=0;
 for(j=0;j<colore;j++) {printf("%d%s",S[j],(j+1)%25?" ":"\n\t ");i+=S[j];}
 printf(" = %d\n",i);
 printf("*******************************************************************************\n");  
 gettimeofday(&ttemp,NULL); 
 getrusage(RUSAGE_SELF,&tempo);
 tot_sec=ttemp.tv_sec-tot_sec0;
 prg_sec= tempo.ru_utime.tv_sec-prg_sec0;
 sys_sec= tempo.ru_stime.tv_sec-sys_sec0;
 tot_microsec=ttemp.tv_usec-tot_microsec0;
 prg_microsec=tempo.ru_utime.tv_usec-prg_microsec0;
 sys_microsec=tempo.ru_stime.tv_usec-sys_microsec0;
 printf("Total Time: %5.3f sec \tCPU Process Time: %5.3f sec \tCPU System Time: %5.3f sec\n",\
     tot_sec+(tot_microsec/1E6),prg_sec+(prg_microsec/1E6),sys_sec+(sys_microsec/1E6));
 //***********************************************************************
 for(j=0;j<(strlen(infile)-3);j++) outfile[j]=infile[j];outfile[j]='\0';
 strncat(outfile,"slt",3);
 ins='a';
 if((file_out=fopen(outfile, "r"))!=NULL)
  {printf("\nThe %s solution file already exists!\nOverwrite (default) or Skip?[ /s]",outfile);
   scanf("%c",&ins);
   fclose(file_out);
  } 
 if(ins!='s')
  {file_out=fopen(outfile, "w");
   if(file_out==NULL)
    {printf("Error in the file opening! Code: %s", strerror(errno));exit(EXIT_FAILURE);}
   printf("Minimum coloring write in the %s file.\n",outfile); 
   //************************************************************************//
    ordina(L,n,0);
   //************************************************************************//
   for(j=n-1;j>-1;j--)
    fprintf(file_out,"%d\n",L[j][2]);
   fclose(file_out);
  } 
 //***************************************************************************
 for(j=0;j<n;j++) {free(T[j].succ);free(L[j]);}
 free(T);free(L);free(S);
 return EXIT_SUCCESS;
}
